#
# steering, Programme/schedule management for AURA
#
# Copyright (C) 2017, Ingo Leindecker
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import datetime, time, timedelta

from rest_framework.exceptions import ValidationError

from django.core.exceptions import ObjectDoesNotExist
from django.forms.models import model_to_dict
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from program.models import Note, RRule, Schedule, ScheduleConflictError, Show, TimeSlot
from program.utils import parse_date, parse_datetime, parse_time
from steering.settings import (
    AUTO_SET_LAST_DATE_TO_DAYS_IN_FUTURE,
    AUTO_SET_LAST_DATE_TO_END_OF_YEAR,
)


# TODO: add type annotations
def resolve_conflicts(data, schedule_pk, show_pk):
    """
    Resolves conflicts
    Expects JSON POST/PUT data from /shows/1/schedules/

    Returns a list of dicts if errors were found
    Returns an empty list if resolution was successful
    """

    sdl = data["schedule"]
    solutions = data.get("solutions", [])

    # Regenerate conflicts
    schedule = instantiate_upcoming(sdl, show_pk, schedule_pk)
    show = schedule.show
    conflicts = make_conflicts(sdl, schedule_pk, show_pk)

    if schedule.rrule.freq > 0 and schedule.first_date == schedule.last_date:
        raise ValidationError(
            _("Start and end dates must not be the same."),
            code="no-same-day-start-and-end",
        )

    if schedule.last_date < schedule.first_date:
        raise ValidationError(
            _("End date mustn't be before start."),
            code="no-start-after-end",
        )

    num_conflicts = len([pr for pr in conflicts["projected"] if len(pr["collisions"]) > 0])

    if len(solutions) != num_conflicts:
        raise ScheduleConflictError(
            _("Numbers of conflicts and solutions don't match."),
            code="one-solution-per-conflict",
            conflicts=conflicts,
        )

    # Projected timeslots to create
    create = []

    # Existing timeslots to update
    update = []

    # Existing timeslots to delete
    delete = []

    # Error messages
    errors = {}

    for ts in conflicts["projected"]:
        # If no solution necessary
        #
        #     - Create the projected timeslot and skip
        #
        if "solution_choices" not in ts or len(ts["collisions"]) < 1:
            projected_ts = TimeSlot.objects.instantiate(ts["start"], ts["end"], schedule, show)
            create.append(projected_ts)
            continue

        # Check hash (if start, end, rrule or by_weekday changed)
        if not ts["hash"] in solutions:
            errors[ts["hash"]] = _("This change on the timeslot is not allowed.")
            continue

        # If no resolution given
        #
        #     - Skip
        #
        if solutions[ts["hash"]] == "":
            errors[ts["hash"]] = _("No solution given.")
            continue

        # If resolution is not accepted for this conflict
        #
        #     - Skip
        #
        if not solutions[ts["hash"]] in ts["solution_choices"]:
            errors[ts["hash"]] = _("Given solution is not accepted for this conflict.")
            continue

        """Conflict resolution"""

        existing = ts["collisions"][0]
        solution = solutions[ts["hash"]]

        # theirs
        #
        #     - Discard the projected timeslot
        #     - Keep the existing collision(s)
        #
        if solution == "theirs":
            continue

        # ours
        #
        #     - Create the projected timeslot
        #     - Delete the existing collision(s)
        #
        if solution == "ours":
            projected_ts = TimeSlot.objects.instantiate(ts["start"], ts["end"], schedule, show)
            create.append(projected_ts)

            # Delete collision(s)
            for ex in ts["collisions"]:
                try:
                    existing_ts = TimeSlot.objects.get(pk=ex["id"])
                    delete.append(existing_ts)
                except ObjectDoesNotExist:
                    pass

        # theirs-end
        #
        #     - Keep the existing timeslot
        #     - Create projected with end of existing start
        #
        if solution == "theirs-end":
            projected_ts = TimeSlot.objects.instantiate(
                ts["start"], existing["start"], schedule, show
            )
            create.append(projected_ts)

        # ours-end
        #
        #     - Create the projected timeslot
        #     - Change the start of the existing collision to projected end
        #
        if solution == "ours-end":
            projected_ts = TimeSlot.objects.instantiate(ts["start"], ts["end"], schedule, show)
            create.append(projected_ts)

            existing_ts = TimeSlot.objects.get(pk=existing["id"])
            existing_ts.start = parse_datetime(ts["end"])
            update.append(existing_ts)

        # theirs-start
        #
        #     - Keep existing
        #     - Create projected with start time of existing end
        #
        if solution == "theirs-start":
            projected_ts = TimeSlot.objects.instantiate(existing["end"], ts["end"], schedule, show)
            create.append(projected_ts)

        # ours-start
        #
        #     - Create the projected timeslot
        #     - Change end of existing to projected start
        #
        if solution == "ours-start":
            projected_ts = TimeSlot.objects.instantiate(ts["start"], ts["end"], schedule, show)
            create.append(projected_ts)

            existing_ts = TimeSlot.objects.get(pk=existing["id"])
            existing_ts.end = parse_datetime(ts["start"])
            update.append(existing_ts)

        # theirs-both
        #
        #     - Keep existing
        #     - Create two projected timeslots with end of existing start and start of existing
        #       end
        #
        if solution == "theirs-both":
            projected_ts = TimeSlot.objects.instantiate(
                ts["start"], existing["start"], schedule, show
            )
            create.append(projected_ts)

            projected_ts = TimeSlot.objects.instantiate(existing["end"], ts["end"], schedule, show)
            create.append(projected_ts)

        # ours-both
        #
        #     - Create projected
        #     - Split existing into two:
        #       - Set existing end time to projected start
        #       - Create another one with start = projected end and end = existing end
        #
        if solution == "ours-both":
            projected_ts = TimeSlot.objects.instantiate(ts["start"], ts["end"], schedule, show)
            create.append(projected_ts)

            existing_ts = TimeSlot.objects.get(pk=existing["id"])
            existing_ts.end = parse_datetime(ts["start"])
            update.append(existing_ts)

            projected_ts = TimeSlot.objects.instantiate(ts["end"], existing["end"], schedule, show)
            create.append(projected_ts)

    # If there were any errors, don't make any db changes yet
    # but add error messages and return already chosen solutions
    if len(errors) > 0:
        conflicts = Schedule.make_conflicts(model_to_dict(schedule), schedule.pk, show.pk)

        partly_resolved = conflicts["projected"]
        saved_solutions = {}

        # Add already chosen resolutions and error message to conflict
        for index, c in enumerate(conflicts["projected"]):
            # The element should only exist if there was a collision
            if len(c["collisions"]) > 0:
                saved_solutions[c["hash"]] = ""

            if c["hash"] in solutions and solutions[c["hash"]] in c["solution_choices"]:
                saved_solutions[c["hash"]] = solutions[c["hash"]]

            if c["hash"] in errors:
                partly_resolved[index]["error"] = errors[c["hash"]]

        # Re-insert post data
        conflicts["projected"] = partly_resolved
        conflicts["solutions"] = saved_solutions
        conflicts["notes"] = data.get("notes")
        conflicts["playlists"] = data.get("playlists")

        raise ScheduleConflictError(
            _("Not all conflicts have been resolved."),
            code="unresolved-conflicts",
            conflicts=conflicts,
        )

    # Collect upcoming timeslots to delete which might still remain
    del_timeslots = TimeSlot.objects.filter(
        schedule=schedule,
        start__gt=timezone.make_aware(datetime.combine(schedule.last_date, time(0, 0))),
    )
    for del_ts in del_timeslots:
        delete.append(del_ts)

    # If 'dryrun' is true, just return the projected changes instead of executing them
    if "dryrun" in sdl and sdl["dryrun"]:
        return {
            "create": [model_to_dict(ts) for ts in create],
            "update": [model_to_dict(ts) for ts in update],
            "delete": [model_to_dict(ts) for ts in delete],
        }

    """Database changes if no errors found"""

    # Only save schedule if timeslots were created
    if create:
        # Create or update schedule
        schedule.save()

    # Update timeslots
    for ts in update:
        ts.save(update_fields=["start", "end"])

    # Create timeslots
    for ts in create:
        ts.schedule = schedule

        # Reassign playlists
        if "playlists" in data and ts.hash in data["playlists"]:
            ts.playlist_id = int(data["playlists"][ts.hash])

        ts.save()

        # Reassign notes
        if "notes" in data and ts.hash in data["notes"]:
            try:
                note = Note.objects.get(pk=int(data["notes"][ts.hash]))
                note.timeslot_id = ts.id
                note.save(update_fields=["timeslot_id"])

                timeslot = TimeSlot.objects.get(pk=ts.id)
                timeslot.note_id = note.id
                timeslot.save(update_fields=["note_id"])
            except ObjectDoesNotExist:
                pass

    # Delete manually resolved timeslots and those after until
    for dl in delete:
        dl.delete()

    return model_to_dict(schedule)


# TODO: add type annotations
def instantiate_upcoming(sdl, show_pk, pk=None):
    """Returns an upcoming schedule instance for conflict resolution"""
    pk = int(pk) if pk is not None else None
    rrule = RRule.objects.get(pk=int(sdl["rrule"]))
    show = Show.objects.get(pk=int(show_pk))

    is_repetition = True if sdl.get("is_repetition") is True else False
    default_playlist_id = (
        int(sdl["default_playlist_id"]) if sdl.get("default_playlist_id") else None
    )
    add_days_no = int(sdl["add_days_no"]) if sdl.get("add_days_no") else None
    add_business_days_only = True if sdl.get("add_business_days_only") is True else False

    first_date = parse_date(str(sdl["first_date"]))
    start_time = (
        sdl["start_time"] + ":00" if len(str(sdl["start_time"])) == 5 else sdl["start_time"]
    )
    end_time = sdl["end_time"] + ":00" if len(str(sdl["end_time"])) == 5 else sdl["end_time"]

    start_time = parse_time(str(start_time))
    end_time = parse_time(str(end_time))

    if sdl["last_date"]:
        last_date = parse_date(str(sdl["last_date"]))
    else:
        # If last_date was not set, set it to the end of the year or add x days
        if AUTO_SET_LAST_DATE_TO_END_OF_YEAR:
            year = timezone.now().year
            last_date = parse_date(f"{year}-12-31")
        else:
            last_date = first_date + timedelta(days=+AUTO_SET_LAST_DATE_TO_DAYS_IN_FUTURE)

    schedule = Schedule(
        pk=pk,
        by_weekday=sdl["by_weekday"],
        rrule=rrule,
        first_date=first_date,
        start_time=start_time,
        end_time=end_time,
        last_date=last_date,
        is_repetition=is_repetition,
        default_playlist_id=default_playlist_id,
        show=show,
        add_days_no=add_days_no,
        add_business_days_only=add_business_days_only,
    )

    return schedule


# TODO: add type annotations
def make_conflicts(sdl, schedule_pk, show_pk):
    """
    Retrieves POST vars
    Generates a schedule
    Generates conflicts: Returns timeslots, collisions, solutions as JSON
    Returns conflicts dict
    """

    # Generate schedule to be saved
    schedule = instantiate_upcoming(sdl, show_pk, schedule_pk)

    # Copy if first_date changes for generating timeslots
    gen_schedule = schedule

    # Generate timeslots

    # If extending: Get last timeslot and start generating from that date on
    if schedule_pk is not None:
        existing_schedule = Schedule.objects.get(pk=int(schedule_pk))

        if schedule.last_date > existing_schedule.last_date:
            last_timeslot = (
                TimeSlot.objects.filter(schedule=existing_schedule).order_by("start").reverse()[0]
            )
            gen_schedule.first_date = last_timeslot.start.date() + timedelta(days=1)

    timeslots = Schedule.generate_timeslots(gen_schedule)

    # Generate conflicts and add schedule
    conflicts = Schedule.generate_conflicts(timeslots)
    conflicts["schedule"] = model_to_dict(schedule)

    return conflicts
